#include <Wire.h>
#include "MPU6050.h"
#include "arduinoFFT.h"
#include "MAX30105.h"
#include "spo2_algorithm.h"
#include <U8g2lib.h>
#include <BluetoothSerial.h>
#include <string.h>

BluetoothSerial BTSerial;
MAX30105 particleSensor;
MPU6050 mpu;
U8G2_SSD1306_128X64_NONAME_F_HW_I2C u8g2(U8G2_R0, /* reset=*/ U8X8_PIN_NONE);
ArduinoFFT<double> FFT = ArduinoFFT<double>();

#define MAX_BRIGHTNESS 255
#define SAMPLES 128           // Must be a power of 2
#define SAMPLING_FREQ 1000    // Hz

static const unsigned char image_heart_bits[] U8X8_PROGMEM = {0x1c,0x1c,0x3e,0x3e,0x7e,0x3f,0xff,0x7f,0xff,0x7f,0xff,0x7f,0xff,0x7f,0xff,0x7f,0xff,0x3f,0xfe,0x3f,0xfc,0x1f,0xfc,0x1f,0xf8,0x0f,0xf0,0x07,0xe0,0x03,0xc0,0x01};
static const unsigned char image_menu_information_sign_bits[] U8X8_PROGMEM = {0xe0,0x03,0x18,0x0c,0x04,0x10,0x02,0x20,0x02,0x20,0x01,0x40,0x01,0x40,0x01,0x40,0x01,0x40,0x01,0x40,0x02,0x20,0x02,0x20,0x04,0x10,0x18,0x0c,0xe0,0x03};
static uint32_t irBuffer[100]; //infrared LED sensor data
static uint32_t redBuffer[100];  //red LED sensor data
static double realData[SAMPLES];     // Acceleration data for FFT
static double imagData[SAMPLES];     // Imaginary part (set to zero)
static double vibrationData[SAMPLES];
int32_t bufferLength; //data length
int32_t spo2; //SPO2 value
int8_t validSPO2; //indicator to show if the SPO2 calculation is valid
int32_t heartRate; //heart rate value
int8_t validHeartRate;
static char remi[50]; 
static char medtime[50]; // Define a character array for storing received data
char dataToSend[50];// Raw acceleration for magnitude calculation



void setup() {
    Serial.begin(115200);
    Wire.begin(25,26);
    u8g2.begin();  // Initialize the display
    u8g2.enableUTF8Print(); // Enable UTF-8 support
    mpu.initialize();
    BTSerial.begin("MediPulse..");
    Serial.println("bluetooth started");
    if (!mpu.testConnection()) {
        Serial.println("MPU6050 not connected!");
        while (1);
    }
    Serial.println("MPU6050 Connected.");
    if (!particleSensor.begin(Wire, I2C_SPEED_FAST)) //Use default I2C port, 400kHz speed
  {
    Serial.println(F("MAX30105 was not found. Please check wiring/power."));
    while (1);
  }

 

  byte ledBrightness = 60; //Options: 0=Off to 255=50mA
  byte sampleAverage = 4; //Options: 1, 2, 4, 8, 16, 32
  byte ledMode = 2; //Options: 1 = Red only, 2 = Red + IR, 3 = Red + IR + Green
  byte sampleRate = 100; //Options: 50, 100, 200, 400, 800, 1000, 1600, 3200
  int pulseWidth = 411; //Options: 69, 118, 215, 411
  int adcRange = 4096; //Options: 2048, 4096, 8192, 16384

  particleSensor.setup(ledBrightness, sampleAverage, ledMode, sampleRate, pulseWidth, adcRange); //Configure sensor with these settings
}

void loop() {
    double sumSquares = 0; // For RMS calculation
    double maxAccel = 0;   // Peak acceleration

    // Collect acceleration data
    for (int i = 0; i < SAMPLES; i++) {
        int16_t ax, ay, az;
        mpu.getAcceleration(&ax, &ay, &az);

        // Convert raw data to m/s²
        double ax_ms2 = (ax / 16384.0) * 9.81;
        double ay_ms2 = (ay / 16384.0) * 9.81;
        double az_ms2 = (az / 16384.0) * 9.81;

        // Remove DC offset (gravity) for vibration analysis
        double accelVib = sqrt(ax_ms2 * ax_ms2 + ay_ms2 * ay_ms2 + (az_ms2 - 9.81) * (az_ms2 - 9.81));

        // Store acceleration for RMS & FFT
        realData[i] = accelVib;
        vibrationData[i] = accelVib;
        imagData[i] = 0; // FFT imaginary part

        // Update sum of squares for RMS
        sumSquares += accelVib * accelVib;

        // Track peak acceleration
        if (accelVib > maxAccel) {
            maxAccel = accelVib;
        }

        delayMicroseconds(1000000 / SAMPLING_FREQ); // Control sampling rate
    }

    // Compute RMS magnitude
    double rmsAccel = sqrt(sumSquares / SAMPLES);

    // Apply window function for FFT (Hamming)
    FFT.windowing(realData, SAMPLES, FFT_WIN_TYP_HAMMING, FFT_FORWARD);

    // Perform FFT
    FFT.compute(realData, imagData, SAMPLES, FFT_FORWARD);

    // Convert to magnitude
    FFT.complexToMagnitude(realData, imagData, SAMPLES);

    // Find peak frequency
    double peakFreq = FFT.majorPeak(realData, SAMPLES, SAMPLING_FREQ);

    // Display results
    Serial.print("Peak Vibration Acceleration: ");
    Serial.print(maxAccel);
    Serial.println(" m/s²");

    Serial.print("RMS Vibration Acceleration: ");
    Serial.print(rmsAccel);
    Serial.println(" m/s²");

    Serial.print("Dominant Vibration Frequency: ");
    Serial.print(peakFreq);
    Serial.println(" Hz");

      bufferLength = 100; //buffer length of 100 stores 4 seconds of samples running at 25sps

  //read the first 100 samples, and determine the signal range
  for (byte i = 0 ; i < bufferLength ; i++)
  {
    while (particleSensor.available() == false) //do we have new data?
      particleSensor.check(); //Check the sensor for new data

    redBuffer[i] = particleSensor.getRed();
    irBuffer[i] = particleSensor.getIR();
    particleSensor.nextSample(); //We're finished with this sample so move to next sample

    Serial.print(F("red="));
    Serial.print(redBuffer[i], DEC);
    Serial.print(F(", ir="));
    Serial.println(irBuffer[i], DEC);
  }

  //calculate heart rate and SpO2 after first 100 samples (first 4 seconds of samples)
  maxim_heart_rate_and_oxygen_saturation(irBuffer, bufferLength, redBuffer, &spo2, &validSPO2, &heartRate, &validHeartRate);

  //Continuously taking samples from MAX30102.  Heart rate and SpO2 are calculated every 1 second
 //ping the first 25 sets of samples in the memory and shift the last 75 sets of samples to the top
    for (byte i = 25; i < 100; i++)
    {
      redBuffer[i - 25] = redBuffer[i];
      irBuffer[i - 25] = irBuffer[i];
    }

    //take 25 sets of samples before calculating the heart rate.
    for (byte i = 75; i < 100; i++)
    {
      while (particleSensor.available() == false) //do we have new data?
        particleSensor.check(); //Check the sensor for new data

      //digitalWrite(readLED, !digitalRead(readLED)); //Blink onboard LED with every data read

      redBuffer[i] = particleSensor.getRed();
      irBuffer[i] = particleSensor.getIR();
      particleSensor.nextSample(); //We're finished with this sample so move to next sample

      //send samples and calculation result to terminal program through UART
      Serial.print(F("red="));
      Serial.print(redBuffer[i], DEC);
      Serial.print(F(", ir="));
      Serial.print(irBuffer[i], DEC);

      Serial.print(F(", HR="));
      Serial.print(heartRate, DEC);

      Serial.print(F(", HRvalid="));
      Serial.print(validHeartRate, DEC);

      Serial.print(F(", SPO2="));
      Serial.print(spo2, DEC);

      Serial.print(F(", SPO2Valid="));
      Serial.println(validSPO2, DEC);
    }

    //After gathering 25 new samples recalculate HR and SP02
    maxim_heart_rate_and_oxygen_saturation(irBuffer, bufferLength, redBuffer, &spo2, &validSPO2, &heartRate, &validHeartRate);
  
    sprintf(dataToSend, "%d,%d,%.2f", heartRate, spo2, maxAccel);
    BTSerial.println(dataToSend);
    Serial.println("data send..");
    if (BTSerial.available()) {
        int i = 0;
        // Read up to 19 characters, leaving space for the null terminator
        while (BTSerial.available() && i < 19) {
            remi[i] = BTSerial.read();
            i++;
        }
        remi[i] = '\0'; // Null-terminate the string

        // Print the received string
        Serial.print("Received: ");
        Serial.println(remi);
        if (strcmp(remi, "MEDICATION REMINDER") == 0){

            u8g2.clearBuffer();
            u8g2.setFontMode(1);
            u8g2.setBitmapMode(1);
            u8g2.setFont(u8g2_font_timR14_tr);
            u8g2.drawStr(7, 36, "TAKE MEDS");
            u8g2.sendBuffer();
           
            for (int i = 0; i < 10; i++) { // Ensure i increments
            u8g2.drawBox(0, 0, 128, 62);
            delay(500);
            u8g2.sendBuffer();
            u8g2.clearBuffer();
            u8g2.setFont(u8g2_font_timR14_tr);
            u8g2.drawStr(7, 36, "TAKE MEDS");
            u8g2.sendBuffer();
            delay(500);
        }

            
            


        }
        else{
          strncpy(medtime,remi,sizeof(medtime)-1);
          medtime[sizeof(medtime) - 1] = '\0';
;
        }
    }

  

    // Start Drawing
    u8g2.clearBuffer();
    
    // Display Heart Rate with Icon 
    u8g2.drawXBMP(6, 6, 15, 16, image_heart_bits);    u8g2.setFont(u8g2_font_ncenB12_tr);
   u8g2.setFont(u8g2_font_5x7_tr);
char bpm[4];
sprintf(bpm,"%d",heartRate);
char sp[4];
sprintf(sp,"%d",spo2);
char trem[5];
dtostrf(maxAccel,6,2,trem);
u8g2.drawStr(23, 17, bpm);

u8g2.drawStr(5, 35, "spO2");

u8g2.drawStr(2, 56, "Tremor");

u8g2.drawFrame(32, 27, 57, 12);

u8g2.drawFrame(32, 47, 58, 12);
int y = map(spo2,0,100,2,56);

u8g2.drawBox(33, 28, y, 10);


u8g2.drawBox(33, 48, maxAccel, 11);

u8g2.drawXBMP(59, 3, 15, 15, image_menu_information_sign_bits);

u8g2.drawLine(0, 0, 0, 0);

u8g2.drawLine(66, 4, 66, 16);

u8g2.setFont(u8g2_font_4x6_tr);
u8g2.drawStr(59, 24, "MEDS");

u8g2.setFont(u8g2_font_6x13_tr);
u8g2.drawStr(78, 14, medtime);

u8g2.drawStr(74, 14, ":");


u8g2.drawStr(91, 37, sp);


u8g2.drawStr(91, 56, trem);

u8g2.drawLine(0, 0, 0, 0);

u8g2.drawLine(55, 0, 55, 30);

u8g2.drawStr(38, 17, "BPM");

u8g2.drawStr(102, 37, "%");



    // Battery Indicator
    u8g2.drawFrame(110, 0, 15, 8);
    u8g2.drawBox(110, 0, 7, 8);

    // Send Buffer to Display
    u8g2.sendBuffer();
    
    delay(1000);
}
float KalmanFilter(float newAngle, float newRate, float dt) {
    // Prediction step
    angle += dt * (newRate - bias);
    P[0][0] += dt * (dt*P[1][1] - P[0][1] - P[1][0] + Q_angle);
    P[0][1] -= dt * P[1][1];
    P[1][0] -= dt * P[1][1];
    P[1][1] += Q_bias * dt;

    // Update step
    float S = P[0][0] + R_measure;
    float K[2];
    K[0] = P[0][0] / S;
    K[1] = P[1][0] / S;

    float y = newAngle - angle;
    angle += K[0] * y;
    bias += K[1] * y; 

    P[0][0] -= K[0] * P[0][0];
    P[0][1] -= K[0] * P[0][1];
    P[1][0] -= K[1] * P[0][0];
    P[1][1] -= K[1] * P[0][1];

    return angle;
}

